package com.make.swing;

import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.DateFormat;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.JPanel;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JButton;
import com.make.util.Debug;

/**
 * <p>A StatusWindow itself is a Thread. Its run() method starts the job
 * given during construction and frequently updates the label with the
 * current text.</p>
 * 
 * <p>The job does whatever it has to do and may call setText() from
 * time to time in order to keep the status message current.</p>
 */
public class StatusWindow extends Thread implements ActionListener {
    /***properties***************************/
    
    /**
     * The JFrame to be shown.
     * 
     * @see createFrame
     * @see positionFrame
     */
    protected JFrame frame;
    
    /**
     * A label that gets frequently updated to contain the current
     * value of the text variable.
     */
    protected JLabel label;
    
    /**
     * Contains a run()-method that is supposed to update the text of this
     * status window via the setText()-method. The run() provided by this
     * thread makes sure that the label showing the text gets updated
     * frequently.
     */
    protected Runnable job;
    
    /**
     * The text to be displayed by <code>label</code>. A client class
     * may update this by calling <code>setText()</code>.
     */
    protected String text;
    
    /**
     * The frame's title string.
     */
    protected String title;
    
    /**
     * Number of milliseconds to wait before updateFrame() gets called
     * the first time. This is the minimum time the initial status
     * message is visible.
     */
    protected int wait;
    
    /**
     * Indicates whether the StatusWindow has a cancel button.
     */
    protected boolean hasCancel;
    
    /**
     * The job thread may call the setText() method in order to
     * inform about its progress.
     */
    protected Thread jobThread;
    
    /***Constructor**************************/
    /**
     * Sets the instance variables as given and starts this
     * thread if its not null.
     */
    public StatusWindow(String _text, String _title, int _wait, 
			boolean _cancel, Runnable _job) {
	text = _text; title = _title; wait = _wait; 
	hasCancel = _cancel; job = _job;
	if (job!=null) start();
    }
    /**
     * The same as the other constructor with _job=null, except that
     * the StatusWindow thread gets started here.
     */
    public StatusWindow(String _text, String _title, int _wait, 
			boolean _cancel) {
	text = _text; title = _title; wait = _wait; 
	hasCancel = _cancel;
	start();
    }
    /***Interface**********************/
    /**
     * <p>If this thread is not alive the job is set and this thread
     * is started. Otherwise an InternalError is thrown.</p>
     */
    public void setJob(Runnable _job) {
	if (Debug.DEBUG && isAlive())
	    throw new InternalError();
	job = _job;
	start();
    }
    /**
     * The run() method takes care that texts given here go into
     * the frame's label.
     */
    public void setText(String t) {
	text = t;
    }
	/**
	 * Stops the thread and sets job to <code>null</code>.
	 */
    public void ready() {
	jobThread = null;
	stop();
    }
    /***Internal*****************************/
    /**
     * Creates the frame with the given title and adds a panel containing the
     * text (and a cancel button if desired).
     */
    protected void createFrame() {
	frame = new JFrame(title);
	frame.setDefaultCloseOperation(frame.DO_NOTHING_ON_CLOSE);
	JPanel panel = new JPanel(new BorderLayout(), false);
	panel.setBorder(javax.swing.BorderFactory.createEmptyBorder(10,10,10,10));
	label = new JLabel(text); text = null;
	panel.add("Center", label);
	if (hasCancel) {
	    JButton cancel = new JButton("Abbrechen");
	    cancel.addActionListener(this);
	    panel.add("East", cancel);
	}
	frame.getContentPane().add("Center", panel);
	frame.pack();
    }
    /**
     * Sets the frame size to half screen width * half screen height and shows
     * it.
     */
    protected void positionFrame() {
	Dimension screen = frame.getToolkit().getScreenSize();
	Dimension size = frame.getSize();
	frame.setLocation((screen.width-size.width)/2, 
			  (screen.height-size.height)/2);
	frame.show();
    }
    /**
     * If the text is non-null (that means it has been set via the setText()-
     * method) the label is being updated to show the new text.
     */
    protected void updateFrame() {
	if (text!=null && label!=null) {
	    label.setText(text);
	    text = null;
	}
    }
    /***ActionListener***********************/
    /**
     * Stops the thread. Called when the cancel button is pressed.
     */
    public void actionPerformed(ActionEvent evt) {
		if (isAlive())
		    stop();
    }
    /***Runnable****************************/
    /**
     * Starts the job if it's non-null, waits <code>wait</code>
     * milliseconds, creates and positions the frame, and
     * checks every 250 milliseconds whether the label needs to
     * be updated with a new text.
     */
    public void run() {
	// turn the Runnable into a Thread
	jobThread = job==null?null:new Thread(job);
	try {
	    if (jobThread!=null)
		jobThread.start();
	    
	    if (wait>0) {
		try {
		    if (jobThread==null) {
			Thread.sleep(wait);
		    } else {
			jobThread.join(wait); // wait at most 'wait' milliseconds for this thread to die
		    }
		} catch (InterruptedException e) {
		    e.printStackTrace();
		}
	    }
	    
	    if (jobThread!=null && !jobThread.isAlive()) {
		// jobThread has died in the meantime, i.e. it
		// has finished its task. So we don't need to
		// show a status window.
		return;
	    }
	    createFrame();
	    positionFrame();
	    // wait...
	    while (jobThread==null || jobThread.isAlive()) {
		try {
		    if (jobThread==null)
			Thread.sleep(250);
		    else
			jobThread.join(250);
		} catch (InterruptedException exc) {
		    exc.printStackTrace();
		}
				// update the frame in case the jobThread
				// has called setText() in the meantime
		updateFrame();
	    }
	} finally {
	    if (jobThread!=null && jobThread.isAlive())
		jobThread.stop();
	    if (frame!=null)
		frame.dispose();
	}
    }
    
    /**
     * An example of how you might want to use the StatusWindow class.
     */
    public static void main(String[] args) {
	StatusWindow window = new StatusWindowUsageExample();
    }
}


class StatusWindowUsageExample extends StatusWindow {
	
    StatusWindowUsageExample() {
	super(
	      "This is the StatusWindow!", // initial status message
	      "Window title", // frame's title
	      0, // don't wait
	      false, // no cancel button
	      null // don't start the thread yet
	      );
	
	setJob(new ExampleJobForMain());
    }
    
    /**
     * Illustrates the use of a job for the StatusWindow.
     *
     * @see main
     */
    class ExampleJobForMain implements Runnable {
	/**
	 * Calls StatusWindow's setText(..) every few seconds in
	 * order to update the status message.
	 */
	public void run() {
	    try {
		sleep(3000);
		setText("Here we go...");
		sleep(3000);
		setText("... still there?");
		sleep(3000);
		setText("Bye, Bye!");
		sleep(3000);
	    } catch(InterruptedException exc) {
		exc.printStackTrace();
	    }
	}
    }
}
