package com.make.swing.text;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import javax.swing.text.Position;
import javax.swing.text.BadLocationException;
import javax.swing.text.Segment;
import javax.swing.undo.UndoableEditSupport;
import javax.swing.undo.UndoableEdit;
import javax.swing.event.UndoableEditListener;
import javax.swing.event.UndoableEditEvent;
import com.make.util.Debug;

public abstract class AbstractDocument implements Document {
    /***properties**************************/
    // data
    protected Element root;
    private int inUpdate = 0;
    // undo
    protected boolean inUndo = false;
    protected UndoableEditSupport undoSupport = new UndoableEditSupport(this);
    private DocumentEvent undoEvent;
    // visual
    protected Vector listeners = new Vector();
    /***Constructor****************************/
    public AbstractDocument() {}
    public AbstractDocument(Element _b) {
	setRoot(_b);
    }
    protected void setRoot(Element _r) {
	root = _r;
	if (root!=null) {
	    Element el = root;
	    while (el!=null && !el.isLeaf() && el.getElementCount()==0) {
		Element ch = el.createDefaultChild();
		el.insertChild(ch, 0);
		ch = el;
	    }
	    root.fresh = false;
	}
    }
    /***Undo**********************************/
    public void lockUndo() { 
	inUndo = true; 
	undoEvent = new DocumentEvent(this, true);
    }
    public void unlockUndo() { 
	inUndo = false; 
	if (undoEvent.isValid())
	    fireDocumentChanged(undoEvent);
	undoEvent = null;
    }
    public void beginUpdate() { 
	if (!inUndo)
	    undoSupport.beginUpdate(); 
	inUpdate++;
    }
    public void endUpdate() { 
	inUpdate--;
	if (Debug.DEBUG && inUpdate<0)
	    throw new InternalError("inUpdate is negative");
	if (!inUndo) {
	    undoSupport.endUpdate(); 
	    if (Debug.DEBUG && undoSupport.getUpdateLevel()<0)
		throw new InternalError("updateLevel is negative");
	}
    }
    public final boolean inUndo() { 
	if (Debug.DEBUG && undoSupport.getUpdateLevel()<0)
	    throw new InternalError("updateLevel is negative");
	return undoSupport.getUpdateLevel()>0; 
    }
    public final boolean inUpdate() {
	return inUpdate>0;
    }
    public final void syncReadAccess() {
	if (inUpdate>0) {
	    long timeout = System.currentTimeMillis()+2000;
	    try {
		while (inUpdate>0 && System.currentTimeMillis()<timeout)
		    Thread.sleep(100);
	    } catch (InterruptedException e) {
	    }
	}
    }
    /***Listeners**************************************/
    public void fireDocumentChanged(DocumentEvent event) {
	if (Debug.DEBUG && event!=undoEvent && undoSupport.getUpdateLevel()==0)
	    throw new InternalError("event without undo");
	//event.dump(System.err);
	for (Enumeration enum=listeners.elements(); enum.hasMoreElements();)
	    ((DocumentListener)enum.nextElement()).documentChanged(event);
    }
    /***Document********************************/
    public void addDocumentListener(DocumentListener listener) {
	listeners.addElement(listener);
    }
    public void removeDocumentListener(DocumentListener listener) {
	listeners.removeElement(listener);
    }
    public void addUndoableEditListener(UndoableEditListener listener) {
	undoSupport.addUndoableEditListener(listener);
    }
    public void removeUndoableEditListener(UndoableEditListener listener) {
	undoSupport.removeUndoableEditListener(listener);
    }
    public DocumentEvent getUndoEvent() { 
	return undoEvent; 
    }
    public void postEdit(UndoableEdit e) { 
	if (inUndo) return;
	if (Debug.DEBUG && !inUndo())
	    throw new InternalError("must call beginUpdate before postEdit");
	undoSupport.postEdit(e);
    }
    public void getText(int offset, int length, Segment text) 
	throws BadLocationException {
	//System.err.println("WARNING called deprecated Document::getText");
	if (length<0) length = 0; // TODO: workaround fix
	StringBuffer sb = new StringBuffer(length);
	root.getText(sb, offset, length);
	length = sb.length();
	text.offset = 0;
	text.count = length;
	text.array = new char[length];
	if (length>0)
	    sb.getChars(0, length, text.array, 0);
    }
    public Element getRootElement() {
	return root;
    }
    public abstract Element getParagraphElement(int pos, Position.Bias bias) 
	throws BadLocationException;
    public void remove(int offs, int len) throws BadLocationException {
	beginUpdate();
	try {
	    DocumentEvent evt = new DocumentEvent(this);
	    root.remove(offs, len, evt);
	    if (evt.isValid())
		fireDocumentChanged(evt);
	} finally {
	    endUpdate();
	}
    }
    public void insertString(int offset, String str, String format)
	throws BadLocationException {
	beginUpdate();
	try {
	    DocumentEvent evt = new DocumentEvent(this);
	    if (root.insertString(offset, str, evt) && evt.isValid())
		fireDocumentChanged(evt);
	} finally {
	    endUpdate();
	}
    }
    public abstract Hashtable getStyle(String name);
    /***misc***********************************/
    public void insertChar(int offset, char ch)
	throws BadLocationException {
	beginUpdate();
	try {
	    DocumentEvent evt = new DocumentEvent(this);
	    if (root.insertString(offset, ""+ch, evt) && evt.isValid())
		fireDocumentChanged(evt);
	} finally {
	    endUpdate();
	}
    }
    public void replace(int start, int len, String str) 
	throws BadLocationException {
	try {
	    beginUpdate();
	    if (len>0) remove(start, len);
	    if (str!=null)
		insertString(start, str, null);
	} finally {
	    endUpdate();
	}
    }
    /***Find***************************************/
    public void unformat(String style, int start, int end) 
	throws BadLocationException {
	if (start>=end) return;
	beginUpdate();
	DocumentEvent evt = new DocumentEvent(this);
	root.removeStyle(style, start, end, evt);
	if (evt.isValid())
	    fireDocumentChanged(evt);
	endUpdate();
    }
    protected abstract Element createStyleElement(String style);
    public void format(String style, int start, int end)
	throws BadLocationException {
	if (start>=end) return;
	beginUpdate();
	boolean already = false;
	try {
	    already = root.isDeepStyle(start, Position.Bias.Forward, style) || 
		root.isDeepStyle(end, Position.Bias.Backward, style);
	} catch (BadLocationException e) {
	    e.printStackTrace(); throw new InternalError();
	}
	DocumentEvent evt = new DocumentEvent(this);
	if (already)
	    root.removeStyle(style, start, end, evt);
	else
	    root.embedSplit(createStyleElement(style), start, end, evt);
	if (evt.isValid())
	    fireDocumentChanged(evt);
	endUpdate();
    }
    public void format(Element el, int start, int end) 
	throws BadLocationException {
	beginUpdate();
	DocumentEvent evt = new DocumentEvent(this);
	root.embedSplit(el, start, end, evt);
	if (evt.isValid())
	    fireDocumentChanged(evt);
	endUpdate();
    }
    public boolean pasteElementAt(int pos, Element el) {
	beginUpdate();
	DocumentEvent evt = new DocumentEvent(this);
	try {
	    if (root.pasteElementAt(pos, el, evt) && evt.isValid()) {
		fireDocumentChanged(evt);
		return true;
	    }
	} catch (BadLocationException e) {
	} finally {
	    endUpdate();
	}
	return false;
    }
    public boolean embed(Element parent, int start, int end) {
	if (start>end) return false;
	beginUpdate();
	DocumentEvent evt = new DocumentEvent(this);
	try {
	    if (root.embed(parent, start, end, evt) && evt.isValid()) {
		fireDocumentChanged(evt);
		return true;
	    }
	} catch (BadLocationException e) {
	} finally {
	    endUpdate();
	}
	return false;
    }
}
