package com.make.swing.text;

import java.util.Vector;
import java.util.Enumeration;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Graphics;
import java.awt.event.FocusListener;
import java.awt.event.FocusEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import javax.swing.SwingUtilities;
import javax.swing.Timer;
import javax.swing.Action;
import javax.swing.text.BadLocationException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import com.make.util.Debug;
import com.make.swing.text.view.Highlighter;
import com.make.swing.text.view.DefaultHighlighter;

public class DefaultCaret
    implements Caret, FocusListener, MouseListener, MouseMotionListener {
    /***properties******************************************/
    private TextComponent component;
    protected Vector listeners = new Vector();
    protected Rectangle rectangle = new Rectangle();
    protected ChangeEvent changeEvent = new ChangeEvent(this);
    protected boolean visible;
    protected int dot;
    protected int mark;
    protected Object selectionTag;
    protected Timer timer;
    protected Point magicCaretPosition;
    protected UpdateHandler updateHandler = new UpdateHandler();
    protected Runnable repaintNewCaret = new RepaintNewCaret();
    /***Constructor*************************************/
    /***Info*******************************************/
    protected final TextComponent getComponent() {
	return component;
    }
    /***Paint**************************************/
    protected final synchronized void repaint() {
	if (component!=null)
	    component.repaint(rectangle.x, rectangle.y, 
			      rectangle.width, rectangle.height);
    }
    protected synchronized void damage(Rectangle r) {
	if (r!=null) {
	    rectangle.x = r.x - 4;
	    rectangle.y = r.y;
	    rectangle.width = 10;
	    rectangle.height = r.height;
	    repaint();
	}
    }
    protected void adjustVisibility(Rectangle r) {
	SwingUtilities.invokeLater(new SafeScroller(r));
    }
    protected Highlighter.HighlightPainter getSelectionPainter() {
	return DefaultHighlighter.getDefaultPainter();
    }
    /***Move**************************************/
    protected void positionCaret(MouseEvent e) {
	Point pt = new Point(e.getX(), e.getY());
	int pos = component.getUI().viewToModel(component, pt);
	if (pos>=0) {
	    setDot(pos);
	    setMagicCaretPosition(null);
	}
    }
    protected void moveCaret(MouseEvent e) {
	Point pt = new Point(e.getX(), e.getY());
	int pos = component.getUI().viewToModel(component, pt);
	if (pos>=0)
	    moveDot(pos);
    }
    /***FocusListener***************************************/
    public void focusGained(FocusEvent evt) {
	if (component.isEditable())
	    setVisible(true);
    }
    public void focusLost(FocusEvent evt) {
	setVisible(false);
    }
    /***MouseListener***************************************/
    public void mouseClicked(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {
	if (SwingUtilities.isLeftMouseButton(e)) {
	    positionCaret(e);
	    if (component.isEnabled())
		component.requestFocus();
	}
    }
    public void mouseReleased(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    /***MouseMotionListener**********************************/
    public void mouseDragged(MouseEvent e) {
	if (SwingUtilities.isLeftMouseButton(e))
	    moveCaret(e);
    }
    public void mouseMoved(MouseEvent e) {}
    /***Fire*********************************************/
    protected void fireStateChanged() {
	for (Enumeration enum=listeners.elements(); enum.hasMoreElements();)
	    ((ChangeListener)enum.nextElement()).stateChanged(changeEvent);
    }	
    /***Caret**********************************************/
    public void install(TextComponent c) {
	if (Debug.DEBUG && component!=null)
	    throw new InternalError();
	component = c;
	Document doc = c.getDocument();
	dot = mark = 0;
	if (doc!=null)
	    doc.addDocumentListener(updateHandler);
	c.addPropertyChangeListener(updateHandler);
	c.addFocusListener(this);
	c.addMouseListener(this);
	c.addMouseMotionListener(this);
	if (component.hasFocus())
	    focusGained(null);
    }
    public void deinstall(TextComponent c) {
	if (Debug.DEBUG && component!=c)
	    throw new InternalError();
	if (timer!=null)
	    timer.stop();
	c.removeMouseListener(this);
	c.removeMouseMotionListener(this);
	c.removeFocusListener(this);
	c.removePropertyChangeListener(updateHandler);
	Document doc = c.getDocument();
	if (doc!=null)
	    doc.removeDocumentListener(updateHandler);
	component = null;
    }
    public void addChangeListener(ChangeListener l) {
	listeners.addElement(l);
    }
    public void removeChangeListener(ChangeListener l) {
	listeners.removeElement(l);
    }
    public void setBlinkRate(int rate) {
	if (rate==0) {
	    if (timer!=null) {
		timer.stop();
		timer.removeActionListener(updateHandler);
		timer = null;
	    }
	} else {
	    if (timer==null)
		timer = new Timer(rate, updateHandler);
	    timer.setDelay(rate);
	}
    }
    public int getBlinkRate() {
	return timer==null ? 0 : timer.getDelay();
    }
    public void paint(Graphics g) {
	if (isVisible()) try {
	    TextUI ui = component.getUI();
	    Rectangle r = ui.modelToView(component, dot);
	    g.setColor(component.getCaretColor());
	    g.drawLine(r.x, r.y, r.x, r.y+r.height-1);
	} catch (BadLocationException e) {
	    e.printStackTrace();
	}
    }
    public boolean isVisible() {
        return visible;
    }
    public void setVisible(boolean e) {
	if (component!=null)
	    if (visible!=e) {
		try {
		    TextUI ui = component.getUI();
		    Rectangle loc = ui.modelToView(component, dot);
		    damage(loc);
		} catch (BadLocationException ex) {
		    ex.printStackTrace();
		}
		visible = e;
	    }
	if (timer!=null)
	    if (visible)
		timer.start();
	    else
		timer.stop();
    }
    public boolean isSelectionVisible() {
	return selectionTag!=null;
    }
    public void setSelectionVisible(boolean vis) {
	if (vis) {
	    Highlighter h = component.getHighlighter();
	    if (dot!=mark && h!=null && selectionTag==null) {
		final int p0 = Math.min(dot, mark);
		final int p1 = Math.max(dot, mark);
		Highlighter.HighlightPainter p = getSelectionPainter();
		try {
                    selectionTag = h.addHighlight(p0, p1, p);
		} catch (BadLocationException bl) {
		    selectionTag = null;
		}
	    }
	} else {
	    if (selectionTag!=null) {
		Highlighter h = component.getHighlighter();
		h.removeHighlight(selectionTag);
		selectionTag = null;
	    }
	}
    }
    public void setMagicCaretPosition(Point p) {
	magicCaretPosition = p;
    }
    public Point getMagicCaretPosition() {
	return magicCaretPosition;
    }
    public int getDot() { return dot; }
    public int getMark() { return mark; }
    public void setDot(int _dot) {
	Document doc = component.getDocument();
	if (doc!=null)
	    _dot = Math.min(_dot, doc.getRootElement().getLength());
	_dot = Math.max(_dot, 0);
	mark = _dot;
	if (dot!=_dot || selectionTag!=null)
	    changeCaretPosition(_dot);
	Highlighter h = component.getHighlighter();
	if (h!=null && selectionTag!=null) {
	    h.removeHighlight(selectionTag);
	    selectionTag = null;
	}
    }
    public void moveDot(int dot) {
	if (!component.isEnabled()) {
	    setDot(dot);
	    return;
	}
	if (dot!=this.dot) {
	    changeCaretPosition(dot);
	    Highlighter h = component.getHighlighter();
	    if (h != null) {
		int p0 = Math.min(dot, mark);
		int p1 = Math.max(dot, mark);
		try {
		    if (selectionTag != null) {
                        h.changeHighlight(selectionTag, p0, p1);
		    } else {
			Highlighter.HighlightPainter p = getSelectionPainter();
                        selectionTag = h.addHighlight(p0, p1, p);
		    }
		} catch (BadLocationException e) {
		    throw new InternalError("Bad caret position");
		}
	    }
	}
    }
    /***Local**************************************/
    protected void changeCaretPosition(int _dot) {
	repaint();
	if (timer!=null && timer.isRunning()) {
            visible = true;
            timer.restart();
        }
	dot = _dot;
        fireStateChanged();
	repaintNewCaret();
    }
    protected void repaintNewCaret() {
	if (component!=null)
	    SwingUtilities.invokeLater(repaintNewCaret);
    }
    /***SafeScroller*********************************/
    class SafeScroller implements Runnable {
	/***properties***************/
	Rectangle r;
	/***Constructor*****************/
	SafeScroller(Rectangle _r) {
	    r = _r;
	}
	/***Runnable******************/
	public void run() {
	    if (component!=null)
		component.scrollRectToVisible(r);
	}
    }
    /***RepaintCaret************************************/
    class RepaintNewCaret implements Runnable {
	public void run() {
	    TextUI mapper = (TextUI)component.getUI();
	    Document doc = component.getDocument();
	    if (mapper!=null && doc!=null) {
		Rectangle newLoc;
		try {
		    newLoc = mapper.modelToView(component, dot);
		} catch (BadLocationException e) {
		    newLoc = null;
		}
		if (newLoc!=null)
		    adjustVisibility(newLoc);
		damage(newLoc);
	    }
	}
    }
    /***UpdateHandler***************************************/
    class UpdateHandler implements PropertyChangeListener, 
				   DocumentListener, ActionListener {
	/***ActionListener************************************/
        public void actionPerformed(ActionEvent e) {
	    if (!visible && (rectangle.width==0 || rectangle.height==0))
		setVisible(true);
	    else {
		visible = !visible;
		repaint();
	    }
	}
	/***DocumentListener******************************************/
        public void documentChanged(DocumentEvent evt) {
	    int ins = evt.getInsertedCount();
	    int rem = evt.getRemovedCount();
	    if (ins>0 || rem>0) {
		final int pos = evt.getPosition();
		if (ins==rem)
		    changeCaretPosition(dot);
		else if (ins>rem) {
		    ins -= rem;
		    int newDot=dot, newMark=mark;
		    if (dot>=pos)
			newDot = dot+ins;
		    if (mark>=pos)
			newMark = mark+ins;
		    setDot(newMark);
		    if (newDot!=newMark)
			moveDot(newDot);
		} else {
		    rem -= ins;
		    int newDot=dot, newMark=mark;
		    if (dot>=(pos+rem))
			newDot = dot-rem;
		    else if (dot>=pos)
			newDot = pos;
		    if (mark>=(pos+rem))
			newMark = mark-rem;
		    else if (mark>=pos)
			newMark = pos;
		    if (Debug.DEBUG) {
			Document doc = evt.getDocument();
			Element root = doc.getRootElement();
			if (newDot<0 || newDot>root.getLength())
			    System.err.println("WARNING: dot="+newDot+" len="+root.getLength());
			if (newMark<0 || newMark>root.getLength())
			    System.err.println("WARNING: mark="+newMark+" len="+root.getLength());
		    }
		    // workaround for delete last line
		    final int len = evt.getDocument().getRootElement().getLength();
		    if (newDot==len)
			newDot--;
		    if (newMark==len)
			newMark--;
		    setDot(newMark);
		    if (newDot!=newMark)
			moveDot(newDot);
		}
	    } else
		repaintNewCaret();
	}
	/***PropertyChangeListener********************************/
	public void propertyChange(PropertyChangeEvent evt) {
	    Object oldValue = evt.getOldValue();
	    Object newValue = evt.getNewValue();
	    if (evt.getSource()==component)
		if (evt.getPropertyName().equals(component.DOCUMENT)) {
		    setDot(0);
		    if (oldValue!=null)
			((Document)oldValue).removeDocumentListener(this);
		    if (newValue!=null)
			((Document)newValue).addDocumentListener(this);
		}
	}
    }
}
