package com.make.swing.text.view;

import java.awt.Shape;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Color;
import java.awt.Graphics;
import javax.swing.text.Segment;
import javax.swing.text.BadLocationException;
import com.make.util.Debug;
import com.make.swing.text.Element;
import com.make.swing.text.Document;
import com.make.swing.text.DocumentEvent;
import com.make.swing.text.TextComponent;
import com.make.swing.text.TextUI;

public abstract class LabelView extends View {
    public final class Style {
	/***properties*************************/
	protected Font font;
	protected FontMetrics metrics;
	protected Color fg;
	protected Color bg;
	protected boolean underline, strike, superscript, subscript;
	protected Toolkit toolkit;
	/***Constructor*****************************/
	public Style(Toolkit tk) {
	    toolkit = tk;
	}
	public void loadPropertiesFromAttributes(Document doc, Element attr) {
	    TextComponent target = getContainer();
	    TextUI ui = target.getUI();
	    font = ui.getFont(target, attr);
	    fg = (Color)attr.getAttribute("textcolor");
	    if (fg==null) fg = Color.black;
	    bg = (Color)attr.getAttribute("bgcolor");
	    underline = attr.containsAttribute("underline", Boolean.TRUE);
	    strike = attr.containsAttribute("strikethrough", Boolean.TRUE);
	    superscript = attr.containsAttribute("superscript", Boolean.TRUE);
	    subscript = attr.containsAttribute("subscript", Boolean.TRUE);
	    metrics = toolkit.getFontMetrics(font);
	}
	/***Formatinfo******************************************/
	public int getTextWidth(Segment s, int offset) {
	    if (Debug.DEBUG) {
		if (offset>s.count)
		    throw new InternalError("invalid offset: ofs="+offset+" count="+s.count);
		if ((s.offset+s.count)>s.array.length)
		    throw new InternalError("invalid segment: len="+s.array.length+
					    " ofs="+s.offset+" count="+s.count);
	    }
	    int nextX = 0;
	    final char[] txt = s.array;
	    final int n = s.offset + offset;
	    for (int i=s.offset; i<n; i++)
		nextX += metrics.charWidth(txt[i]);
	    return nextX;
	}
	public int getTextWidth(Segment s) {
	    return getTextWidth(s, s.count);
	}
	public int getTextHeight() {
	    return metrics.getHeight();
	}
	public Color getSelectionColor(TextComponent c) {
	    if (c==null) return null;
	    Color sc = c.getSelectedTextColor();
	    return sc==null || sc.equals(fg)?null:sc;
	}
	public int getFontHeight() { return metrics.getHeight(); }
	public int getFontDescent() { return metrics.getDescent(); }
	/***Search******************************************/
	public int getTextOffset(Segment s, int x, boolean round) {
	    int currX=0, nextX=currX;
	    char[] txt = s.array;
	    final int n = s.offset + s.count;
	    for (int i=s.offset; i<n; i++) {
		nextX += metrics.charWidth(txt[i]);
		if (x>=currX && x<nextX)
		    if (round==false || ((x-currX)<(nextX-x)))
			return i - s.offset;
		    else
			return i + 1 - s.offset;
		currX = nextX;
	    }
	    return s.count;
	}
	public int getTextOffset(Segment s, int x) {
	    return getTextOffset(s, x, true);
	}
	/***Draw***********************************************/
	public int drawText(Segment s, int x, int y, Graphics g) {
	    final int n = s.offset + s.count;
	    if (Debug.DEBUG && n>s.array.length) {
		getDocument().getRootElement().dump(System.err);
		throw new InternalError(LabelView.this+".drawText len="+s.array.length+
					" ofs="+s.offset+" count="+s.count);
	    }
	    if (s.count>0)
		g.drawChars(s.array, s.offset, s.count, x, y);
	    int nextX = x;
	    for (int i=s.offset; i<n; i++)
		nextX += metrics.charWidth(s.array[i]);
	    return nextX;
	}
	public int paintText(Segment s, int x, int y, Graphics g, 
			     Color color, boolean stripWS) {
	    final int startX = x;
	    g.setColor(color);
	    if (superscript)
		y -= metrics.getAscent() / 2;
	    else if (subscript)
		y += metrics.getDescent();
	    x = drawText(s, x, y, g);
	    if (underline || strike) {
		int width = x-startX;
		if (stripWS) {
		    while (s.count>0 && 
			   (Character.isWhitespace(s.array[s.count-1]) || 
			    s.array[s.count-1]=='\u00b7')) {
			width -= metrics.charWidth(s.array[s.count-1]);
			s.count -= 1;
		    }
		}
		if (underline)
		    y += 1;
		else if (strike)
		    y -= (metrics.getAscent()/2);
		g.drawLine(startX, y, x, y);
	    }
	    return x;
	}
	public int paintText(Segment s, int x, int y, Graphics g, 
			     boolean stripWS) {
	    return paintText(s, x, y, g, fg, stripWS);
	}
	public void paintBackground(Graphics g, Shape a) {
	    if (bg!=null) {
		Rectangle alloc = (a instanceof Rectangle)
		    ?(Rectangle)a
		    :a.getBounds();
		g.setColor(bg);
		g.fillRect(alloc.x, alloc.y, alloc.width, alloc.height);
	    }
	}
	/***Break******************************************/
	public int getBreakLocation(Segment s, int x) {
	    final int index = getTextOffset(s, x, false);
	    final int first = s.offset + Math.min(index, s.count-1);
	    for (int i=first; i>=s.offset; i--) {
		final char ch = s.array[i];
		if (Character.isWhitespace(ch) || ch=='\u00b7')
		    return i-s.offset+1;
	    }
	    return index;
	}
	public int getBreakWeight(Segment s, int offset, float pos, float len) {
	    // TODO : offset beachten
	    final int index = getTextOffset(s, (int)len);
	    if (index==0)
		return BadBreakWeight;
	    final int first = s.offset+Math.min(index, s.count-1);
	    for (int i=first; i>=s.offset; i--) {
		final char ch = s.array[i];
		if (Character.isWhitespace(ch) || ch=='\u00b7')
		    return ExcellentBreakWeight;
	    }
	    return GoodBreakWeight;
	}
    }
    /***properties*********************************/
    private int allocX, allocY;
    protected Segment text = new Segment();
    protected Style style = new Style(Toolkit.getDefaultToolkit());
    protected boolean styleMaster = true;
    /***Constructor************************************/
    public LabelView(Element elem) { super(elem); }
    /***LabelView*********************************/
    protected void loadText() {
	text.offset = 0; text.count = text.array.length;
    }
    protected abstract void loadTextFromElement();
    /***Misc***********************************************/
    public Segment getText() { return text; }
    protected final void sync() {
	if (styleMaster && style.font==null)
	    style.loadPropertiesFromAttributes(getDocument(), getElement());
	if (text.array==null)
	    loadTextFromElement();
	loadText();
	if (Debug.DEBUG) {
	    if (text.array==null)
		throw new InternalError();
	    if (text.offset<0)
		throw new InternalError();
	    if ((text.offset+text.count)>text.array.length)
		throw new InternalError("this="+this+" offs="+text.offset+" count="+text.count+" arraylen="+text.array.length);
	}
    }
    public Style getStyle() { return style; }
    public int getTextOffset(int relX, boolean round) {
	return getStartOffset()+style.getTextOffset(text, relX, round);
    }
    /***View*************************************************/
    public View createFragment(int p0, int p1) {
	return new LabelFragment(this, p0, p1);
    }
    public void documentChanged(DocumentEvent e, Shape a, ViewFactory f) {
	text.array = null;
	style.font = null;
	sync();
    }
    public float getMaximumSpan(int axis) {
	return axis==Y_AXIS
	    ?getPreferredSpan(axis)
	    :super.getMaximumSpan(axis);
    }
    public float getPreferredSpan(int axis) {
	sync();
	if (axis==X_AXIS) {
	    final int width = style.getTextWidth(text);
	    return Math.max(width, 1);
	} else
	    return style.getFontHeight();
    }
    public Shape modelToView(int pos, Shape a) throws BadLocationException {
	Rectangle alloc = a instanceof Rectangle
	    ?(Rectangle)a
	    :a.getBounds();
	sync();
	if (pos>=getEndOffset())
	    return new Rectangle(alloc.x+alloc.width, alloc.y, 
				 0, style.getFontHeight());
	int width = style.getTextWidth(text, pos-getStartOffset());
	return new Rectangle(alloc.x+width, alloc.y, 0, style.getFontHeight());
    }
    public int viewToModel(float x, float y, Shape a) {
	Rectangle alloc = a.getBounds();
	sync();
	final int offs = style.getTextOffset(text, (int)x-alloc.x);
	final int retValue = getStartOffset() + offs;
	return retValue;
    }
    public int getNextVisualPositionFrom(int pos, int direction) 
	throws BadLocationException {
	if (DUMP_NEXTVISUAL)
	    System.err.println(this+": nextVisual pos="+pos+" dir="+direction+" end="+getEndOffset());
	switch (direction) {
	case NORTH: 
	case SOUTH:
	    if (pos==-1) {
		sync();
		if (text.count==0) 
		    return getStartOffset();
		final int x = getMagicX()-allocX;
		if (x<0)
		    return getStartOffset();
		else
		    return getTextOffset(x, false);
	    } else
		return -1;
	case EAST:
	    if (pos==-1)
		return getStartOffset();
	    pos++;
	    if (pos>=getEndOffset())
		return -1;
	    if (DUMP_NEXTVISUAL)
		System.err.println(this+": nextMajor res="+pos);
	    return pos;
	case WEST:
	    if (pos==-1)
		return getEndOffset()-1;
	    if (pos<=getStartOffset())
		return -1;
	    return pos-1;
	default:
	    throw new IllegalArgumentException("Bad direction: " + direction);
	}
    }
    final int paintText(int x, int y, Graphics g, int p0,
			int p1, boolean stripWS) {
	sync();
	text.count = p1-p0;
	text.offset += p0-getStartOffset();
	return style.paintText(text, x, y, g, stripWS);
    }
    final int paintText(int x, int y, Graphics g, int p0,
			int p1, Color color, boolean stripWS) {
	sync();
	text.count = p1-p0;
	text.offset += p0-getStartOffset();
	return style.paintText(text, x, y, g, color, stripWS);
    }
    protected void paintText(TextComponent c, Graphics g, Shape a,
			     boolean stripWhitespace) {
	final int p0 = getStartOffset();
	final int p1 = getEndOffset();
	Rectangle alloc = a.getBounds();
	sync();
	int y = alloc.y + alloc.height - style.getFontDescent();
	g.setFont(style.font);
	paintText(alloc.x, y, g, p0, p1, stripWhitespace);
    }
    /***View***************************************************/
    public void paint(Graphics g, Shape a) {
	Rectangle alloc = a.getBounds();
	allocX = alloc.x; allocY = alloc.y;
	style.paintBackground(g, a);
	TextComponent tc = getContainer();
	Highlighter h = tc.getHighlighter();
	h.paintHighlights(g, getStartOffset(), getEndOffset(), 
			  a, tc, this);
	paintText(tc, g, a, false);
    }
    public int getBreakWeight(int axis, int offset, float pos, float len) {
	if (axis==X_AXIS) {
	    loadTextFromElement(); sync();
	    return style.getBreakWeight(text, offset, pos, len);
	} else
	    return super.getBreakWeight(axis, offset, pos, len);
    }
    public View breakView(int axis, int p0, float pos, float len) {
	if (axis==X_AXIS) {
	    sync();
	    final int d = p0 - getStartOffset();
	    text.count -= d;
	    text.offset += d;
	    int p1 = p0+style.getBreakLocation(text, (int)len);
	    return createFragment(p0, p1);
	}
	return this;
    }
    public float getAlignment(int axis) {
	if (axis == Y_AXIS) {
	    float h = style.getFontHeight();
	    float d = style.getFontDescent();
	    float align = (h - d) / h;
	    return align;
	} 
	return super.getAlignment(axis);
    }
}
