package com.make.swing.text.view;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.awt.Shape;
import java.awt.Graphics;
import java.awt.Component;
import java.awt.Rectangle;
import java.awt.Point;
import javax.swing.SizeRequirements;
import javax.swing.SwingConstants;
import javax.swing.text.StyleConstants;
import javax.swing.text.BadLocationException;
import com.make.xml.DocumentType;
import com.make.xml.ElementType;
import com.make.util.Debug;
import com.make.swing.text.DocumentEvent;
import com.make.swing.text.ContainerChange;
import com.make.swing.text.Element;
import com.make.swing.text.TextComponent;

public class ParagraphView extends BoxView implements BreakableView, VerticalView {
    /***properties***********************************/
    private int justification;
    private float lineSpacing;
    protected int firstLineIndent;
    private int layoutSpan;
    protected ViewFactory viewFactory;
    private Vector layoutPool;
    private String prefix;
    protected PrefixView prefixView;
    protected int prefixWidth;
    private int rowPosition = -1;
    /***Constructor**********************************/
    public ParagraphView(Element elem, ViewFactory vf) {
	super(elem, Y_AXIS);
	viewFactory = vf;
	layoutSpan = -1;
	setPropertiesFromAttributes();
    }
    public ViewFactory getViewFactory() { return viewFactory; }
    public float getMaximumSpan(int axis) {
	if (axis==getAxis())
	    return getPreferredSpan(axis);
	else
	    return super.getMaximumSpan(axis);
    }
    /***BreakableView********************************/
    public View getPoolParent() { return this; }
    public int getPoolViewIndexAt(int position) {
	return getElement().getIndexAt(position);
    }
    public View getPoolView(int index) { 
	return (View)layoutPool.elementAt(index); 
    }
    public int getPoolViewCount() { return layoutPool.size(); }
    public View createView(Vector children) {
	Row row = new Row(getElement(), rowPosition, justification);
	for (Enumeration enum=children.elements(); enum.hasMoreElements();)
	    row.append((View)enum.nextElement());
	rowPosition = row.getEndOffset()-getStartOffset();
	return row;
    }
    /***CompositeView************************************/
    protected void loadChildren(ViewFactory f) {
        layoutPool = new Vector();
        Element e = getElement();
	int n = e.getElementCount();
	for (int i = 0; i < n; i++) {
	    View v = f.create(e.getElementAtIndex(i));
	    v.setParent(this);
	    layoutPool.addElement(v);
	}
    }
    /***BoxView******************************************/
    protected void layout(int width, int height) {
	syncPrefix();
        if (layoutSpan!=width) {
            final int oldHeight = height;
            rebuildRows(width);
            final int newHeight = (int)getPreferredSpan(Y_AXIS);
            if (oldHeight!=newHeight)
                getParent().preferenceChanged(this, false, true);
	}
        super.layout(width, height);
    }
    /***ParagraphView***********************************/
    protected void setPrefix(String _p) { prefix = _p; }
    protected String getPrefix() { return prefix; }
    protected void setJustification(int j) { justification = j; }
    protected void setLineSpacing(float ls) { lineSpacing = ls; }
    protected void setFirstLineIndent(float fi) { firstLineIndent = (int) fi; }
    protected void syncPrefix() {
	if (prefix==null) {
	    if (prefixView!=null)
		prefixView = null;
	    prefixWidth = 0;
	} else {
	    if (prefixView==null)
		prefixView = new PrefixView();
	    else
		prefixView.prefixChanged();
	    prefixWidth = Math.max(4, (int)(prefixView.getPreferredSpan(X_AXIS)*1.2));
	}
	setParagraphInsets(getElement());
	if (prefixView!=null)
	    setInsets(getTopInset(), (short)(getLeftInset()+prefixWidth),
		      getBottomInset(), getRightInset());
    }
    protected void setPropertiesFromAttributes() { 
	Element attr = getElement();
	setParagraphInsets(attr);
	//syncPrefix();
    }
    protected synchronized void rebuildRows(int span) {
        layoutSpan = span;
        int p0 = getStartOffset(); 
        int p1 = getEndOffset()-1;
	rowPosition = 0;
        removeAll();
	// reparent
	for (Enumeration enum=layoutPool.elements(); enum.hasMoreElements();)
	    ((View)enum.nextElement()).setParent(this);
        
	boolean firstRow = true;

	Row row = null;
        while (p0<p1) {
            final int old = p0;
	    float indent = firstRow ? Math.min(span-5,firstLineIndent) : 0;
	    row = layoutRow(p0, indent);
            append(row);
            p0 = row.getEndOffset();
            if (Debug.DEBUG && p0<p1 && p0<=old) {
		System.err.println("infinite loop: "+this);
		try {
		    System.err.println("start = "+getStartOffset());
		    System.err.println("end = "+getEndOffset());
		    System.err.println("p0 = "+p0);
		    System.err.println("p1 = "+p1);
		    System.err.println("old = "+old);
		    System.err.println("layoutSpan = "+layoutSpan);
		    System.err.println("row.nviews = "+row.getViewCount());
		    System.err.println("row.view = "+row.getView(0));
		    for (int z=0; z<getElement().getElementCount(); z++) {
			Element el = getElement().getElementAtIndex(z);
			System.err.println("--el "+z+": "+el);
			System.err.println("    start="+el.getStartOffset()
					   +" end="+el.getEndOffset());
		    }
		    for (int z=0; z<row.getViewCount(); z++) {
			View v = row.getView(z);
			System.err.println("--row "+z+": "+v);
			System.err.println("    start="+v.getStartOffset()
					   +" end="+v.getEndOffset());
		    }
		    System.exit(1);
		} catch (Exception e) {
		}
            }
        }
	if (row!=null) {
	    row.virtualLast = true;
	    row.afterLen = 1;
	}
    }
    protected Row layoutRow(int pos, float indent) {
	// calc
        float x = getLeftInset() + indent;
        float spanLeft = layoutSpan - indent;
	// call internalBreakView
	Row row = (Row)BreakUtil.fillView(this, X_AXIS, pos, x, spanLeft);
	// reparent
        int poolSize = layoutPool.size();
        for (int i=0; i<poolSize; i++) {
            View v = getPoolView(i);
            if (v.getParent()==null)
                v.setParent(this);
        }
	// lineSpacing
	if (lineSpacing>1) {
	    float height = row.getPreferredSpan(Y_AXIS);
	    float addition = (height * lineSpacing) - height;
	    if (addition>0)
		row.setLineSpacing((short)addition);
	}
	return row;
    }
    /***update************************************************/
    protected void updateLayout(DocumentEvent evt, ContainerChange chg, Shape a) {
        layoutSpan = Integer.MAX_VALUE;
	super.updateLayout(evt, chg, a);
	if (a!=null && (chg==null || !(chg.getAdded().isValid() ||
				       chg.getRemoved().isValid()))) {
	    Rectangle alloc = prefixView==null?getInsideAllocation(a):a.getBounds();
	    Component host = getContainer();
	    host.repaint(alloc.x, alloc.y, alloc.width, alloc.height);
	}
    }
    /***View*****************************************************************/
    public float getAlignment(int axis) {
        switch (axis) {
        case Y_AXIS:
	    if (getViewCount()!=0) {
		int paragraphSpan = (int)getPreferredSpan(Y_AXIS);
		int rowSpan = (int)getView(0).getPreferredSpan(Y_AXIS);
		return paragraphSpan!=0 ? ((float)rowSpan/2) / paragraphSpan : 0;
	    } else
		return 0.5f;
	case X_AXIS:
	    return 0.5f;
	default:
            throw new IllegalArgumentException("Invalid axis: " + axis);
	}
    }
    public void documentChanged(DocumentEvent evt, Shape a, ViewFactory f) {
	setPropertiesFromAttributes();
	super.documentChanged(evt, a, f);
    }
    public void paint(Graphics g, Shape allocation) {
	super.paint(g, allocation);
	if (prefixView!=null) {
	    Rectangle r = allocation.getBounds();
	    r.y += getTopInset();
	    r.width = getLeftInset();
	    r.height = (int)prefixView.getPreferredSpan(Y_AXIS);
	    prefixView.paint(g, r);
	}
    }
    /***BoxView*******************************************************/
    protected SizeRequirements calculateMinorAxisRequirements(int axis, SizeRequirements r) {
	if (r==null)
	    r = new SizeRequirements();
	float pref = 0;
	final int count = layoutPool.size();
	for (int i=0; i<count; i++)
	    pref += getPoolView(i).getPreferredSpan(axis);
	float insets = (axis == X_AXIS) 
	    ? getLeftInset() + getRightInset() 
	    : getTopInset() + getBottomInset();
	r.minimum = ((int)insets) + 20;
	r.preferred = Math.max(r.minimum, (int)pref);
	r.maximum = Short.MAX_VALUE;
	r.alignment = 0.5f;
	return r;
    }
    /***Pool**************************************************/
    protected void replacePool(int first, int count, View[] views) {
        for (int i=count; i>0; i--) {
	    View v = (View)layoutPool.elementAt(first);
	    v.setParent(null);
            layoutPool.removeElementAt(first);
        }
        for (int i=0; i<views.length; i++) {
	    View v = views[i];
	    v.setParent(this);
            layoutPool.insertElementAt(v, first+i);
        }
    }
    public Shape getPoolViewAllocation(int idx, Shape a) {
	return null;
    }
    /***Row****************************************************************/
    protected class Row extends BoxView {
	/***properties*****************/
	protected int justification;
	protected int start, afterLen;
	/***Constructor*********************/
        public Row(Element elem, int _start, int just) {
            super(elem, X_AXIS);
	    start = _start;
	    justification = just;
	    minorCursorWrap = false;
        }
	final public void setLineSpacing(short bottom) {
	    setInsets(getTopInset(), getLeftInset(), bottom, getRightInset());
	}
	/***View*********************/
	public int getRowStart(int pos) {
	    final int p = super.getRowStart(pos);
	    final int count = getViewCount();
	    if (p!=-1 || count==0) return p;
	    try {
		return getView(0).getNextVisualPositionFrom(-1, EAST);
	    } catch (BadLocationException e) {
		throw new InternalError(""+e);
	    }
	}
	public int getRowEnd(int pos) {
	    final int p = super.getRowStart(pos);
	    final int count = getViewCount();
	    if (p!=-1 || count==0) return p;
	    return getEndOffset()-1;
	}
        public float getAlignment(int axis) {
            if (axis==X_AXIS) {
                switch (justification) {
                case StyleConstants.ALIGN_LEFT:
                    return 0;
                case StyleConstants.ALIGN_RIGHT:
                    return 1;
                case StyleConstants.ALIGN_CENTER:
                case StyleConstants.ALIGN_JUSTIFIED:
                    return 0.5f;
                }
            }
            return super.getAlignment(axis);
        }
        public int getStartOffset() { return ParagraphView.this.getStartOffset()+start; }
        public int getEndOffset() {
	    int c = getViewCount();
	    return (c==0?getStartOffset():getView(c-1).getEndOffset())+afterLen;
	}
	/***CompositeView*****************/
	protected int getViewIndexAtPosition(int pos) {
	    if (pos<getStartOffset() || pos>getEndOffset())
		return -1;
	    final int count = getViewCount();
	    for (int z=count-1; z>0; z--)
		if (pos>=getView(z).getStartOffset())
		    return z;
	    return 0;
	}
        protected void loadChildren(ViewFactory f) {}
    }
    /***Prefix*********************************************************/
    protected class PrefixView extends LabelView {
	/***Constructor************************/
	public PrefixView() {
	    super(ParagraphView.this.getElement());
	}
	/***View******************************/
	public int getStartOffset() { return ParagraphView.this.getStartOffset(); }
	public int getEndOffset() { return getStartOffset(); }
	/***LabelView**************************/
	protected void loadTextFromElement() {
	    text.array = getPrefix().toCharArray();
	}
	protected void paintText(TextComponent c, Graphics g, Shape a,
				 boolean stripWhitespace) {
	    final int p0 = getStartOffset();
	    final int p1 = p0+text.array.length;
	    Rectangle alloc = a.getBounds();
	    sync();
	    int y = alloc.y + alloc.height - style.getFontDescent();
	    g.setFont(style.font);
	    paintText(alloc.x, y, g, p0, p1, stripWhitespace);
	}
	/***View***************************/
	public TextComponent getContainer() {
	    return ParagraphView.this.getContainer();
	}
	/***internal******************************/
	protected void prefixChanged() {
	    text.array = null;
	}
    }
}
