package com.make.xml;

import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.io.UnsupportedEncodingException;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import com.make.util.Strings;

public class DOMWriter {
    /***static*******************************************/
    public static final int WIDTH = 79;
    public static final String encoding = "ISO-8859-1";
    /***properties***************************************/
    protected PrintWriter out;
    protected DocumentType docType;
    protected boolean spaceAllowed = true;
    protected boolean cfuml = false;
    protected int left = WIDTH;
    protected StringBuffer indent = new StringBuffer();
    /***Constructor**************************************/
    public DOMWriter(PrintWriter wr) {
	out = wr;
    }
    public DOMWriter(Writer wr) {
	this(new PrintWriter(wr));
    }
    public DOMWriter(OutputStream os) throws UnsupportedEncodingException {
	this(new OutputStreamWriter(os, encoding));
    }
    public DOMWriter(OutputStream os, boolean _cfuml) 
	throws UnsupportedEncodingException {
	this(new OutputStreamWriter(os, encoding));
	cfuml = _cfuml;
    }
    public void flush() {
	out.flush();
    }
    /***Props*******************************************/
    public void setDocumentType(DocumentType _dt) {
	docType = _dt;
    }
    /***Print*******************************************/
    protected void printNoBreak(String s) {
	out.print(s); 
	left -= s.length();
    }
    protected void print(String s) {
	if (s.length()>left && spaceAllowed)
	    println();
	printNoBreak(s);
    }
    protected void println() {
	if (!cfuml) {
	    out.println();
	    out.print(indent.toString());
	    left = WIDTH-indent.length();
	}
    }
    public void print(Node node) {
	if (node==null) return;
	int type = node.getNodeType();
	switch (type) {
	case Node.DOCUMENT_NODE: { // print document
	    Document doc = (Document)node;
	    String tag = doc.getDocumentElement().getTagName();
	    indent.setLength(0);
	    if (!cfuml) {
		out.println("<?xml version=\"1.0\" encoding=\""+encoding+"\"?>");
		println();
		/*if (docType!=null) {
		  String systemId = docType.getSystemId();
		  if (systemId!=null) {
		  out.print("<!DOCTYPE "+tag);
		  String publicId = docType.getPublicId();
		  if (publicId==null)
		  out.print(" SYSTEM \"");
		  else
		  out.print(" PUBLIC \""+Strings.escapeXMLText(publicId)+"\" \"");
		  out.print(Strings.escapeXMLText(systemId));
		  out.print('\"');
		  out.println('>');
		  }
		  }*/
	    }
	    //println();
	    NodeList nl = node.getChildNodes();
	    for (int z=0; z<nl.getLength(); z++)
		print(nl.item(z));
	    println();
	    flush();
	    spaceAllowed = true;
	    break;
	}
	
	// print element with attributes
	case Node.ELEMENT_NODE: {
	    ElementType et = docType==null
		?null
		:docType.getElementType(node.getNodeName());
	    final boolean flowing = et==null || et.isFlowing();
	    if (!flowing)
		println();
	    printNoBreak("<");
	    printNoBreak(node.getNodeName());
	    NodeList nl = node.getChildNodes();
	    Node[] children = new Node[nl.getLength()+1];
	    for (int i=0,z=0; i<nl.getLength(); i++) {
		Node ch = nl.item(i);
		if (et!=null) {
		    if (ch instanceof Text &&
			!et.mayContainText())
			continue;
		}
		children[z++] = ch;
	    }
	    Attr attrs[] = sortAttributes(node.getAttributes());
	    for (int i=0; i<attrs.length; i++) {
		Attr attr = attrs[i];
		String val = Strings.escapeXMLText(attr.getNodeValue());
		print(' '+attr.getNodeName()+"=\""+val+'"');
	    }
	    if (children[0]==null) {
		printNoBreak("/>");
	    } else {
		boolean text1 = children[0].getNodeType()==Node.TEXT_NODE &&
		    children[1]==null &&
		    (left-4)>(children[0].getNodeValue().length()+
			      node.getNodeName().length());
		printNoBreak(">");
		if (!flowing) {
		    indent.append("  ");
		    if (!text1 &&
			children[0].getNodeType()!=Node.ELEMENT_NODE)
			println();
		}
		for (int z=0; children[z]!=null; z++)
		    print(children[z]);
		if (!flowing) {
		    indent.setLength(indent.length()-2);
		    if (!text1)
			println();
		}
		print("</"+node.getNodeName()+">");
	    }
	    break;
	}
	
	// handle entity reference nodes
	case Node.ENTITY_REFERENCE_NODE: {
	    print('&'+node.getNodeName()+';');
	    break;
	}
	
	// print cdata sections
	case Node.CDATA_SECTION_NODE: {
	    print("<![CDATA["+node.getNodeValue()+"]]>");
	    break;
	}

	// print text
	case Node.TEXT_NODE: {
	    StringBuffer word = new StringBuffer();
	    String text = cfuml
		?Strings.escapeWDDXText(node.getNodeValue())
		:Strings.escapeXMLText(node.getNodeValue());
	    boolean wasSpace = false;
	    spaceAllowed = true;
	    final char[] textc = text.toCharArray();
	    for (int z=0; z<textc.length; z++) {
		final char ch = textc[z];
		if (Character.isWhitespace(ch)) {
		    if (!wasSpace) {
			word.append(' ');
			print(word.toString());
			word.setLength(0);
			wasSpace = true;
		    }
		} else {
		    word.append(ch);
		    wasSpace = false;
		}
	    }
	    if (word.length()>0)
		print(word.toString());
	    spaceAllowed = wasSpace;
	    break;
	}
	
	// print processing instruction
	case Node.PROCESSING_INSTRUCTION_NODE: {
	    printNoBreak("<?"+node.getNodeName());
	    String data = node.getNodeValue();
	    if (data!=null && data.length()>0)
		printNoBreak(' '+data);
	    printNoBreak("?>");
	    break;
	}
	}
    }
    protected Attr[] sortAttributes(NamedNodeMap attrs) {
	int len = attrs==null ? 0 : attrs.getLength();
	Attr array[] = new Attr[len];
	for (int i=0; i<len; i++) {
	    array[i] = (Attr)attrs.item(i);
	}
	for (int i=0; i<len-1; i++) {
	    String name  = array[i].getNodeName();
	    int index = i;
	    for (int j=i+1; j<len; j++) {
		String curName = array[j].getNodeName();
		if (curName.compareTo(name)<0) {
		    name  = curName;
		    index = j;
		}
	    }
	    if (index!=i) {
		Attr temp = array[i];
		array[i] = array[index];
		array[index] = temp;
	    }
	}
	return array;
    }
    /***Debug******************************/
    public static void main(String[] args) throws Exception {
	new DOMWriter(System.out).print(DOMUtil.parse(System.in));
    }
} 
